abbrmap = {
    "ahu": ["ahu"],
    "avg": ["average"],
    "cmd": ["command"],
    "elec": ["electrical"],
    "equip": ["equipment"],
    "freq": ["frequency"],
    "occ": ["occupied"],
    "rtu": ["rtu"],
    "roof": ["rooftop"],
    "sp": ["setpoint"],
    "spt": ["setpoint"],
    "stpt": ["setpoint"],
    "temp": ["temperature"],
    "unocc": ["unoccupied"],
    "volt": ["voltage"],
    "ctl": ["control"],
    "sa": ["supply", "air"],
    "ea": ["exhaust", "air"],
    "da": ["discharge", "air"],
    "hw": ["hot", "water"],
    "chw": ["chilled", "water"],
    "z": ["zone"],
    "zn": ["zone"],
}
